import numpy as np
import torch
from tqdm import tqdm
from torch.autograd import Variable
from matplotlib import pyplot as plt

class LRFinder:
    def __init__(self,model,train_dataloader,optimizer,device,initial_lr=1e-8,final_lr=10,beta=0.98):
        self.model=model
        self.train_dataloader=train_dataloader
        self.optimizer=optimizer
        self.beta=beta
        self.initial_lr=initial_lr
        self.final_lr=final_lr
        self.device=device
        self.optimizer.param_groups[0]['lr'] = self.initial_lr
        self.num = len(self.train_dataloader)-1
        self.mult = (self.final_lr / self.initial_lr) ** (1/self.num)
        self.loss_list=[]
        self.log_lrs=[]
    def find(self):
        best_loss=np.inf
        avg_loss=0.0
        train=self.train_dataloader
        tk = tqdm(train, total=len(train), position=0, leave=True)
        smoothed_loss=0.0
        for i,data in enumerate(tk):
            for key, value in data.items():
                data[key] = Variable(value.to(self.device))
            self.model.train()
            self.optimizer.zero_grad()
            _, loss = self.model(**data)
            avg_loss = self.beta * avg_loss + (1-self.beta) *loss.data.item()
            #Compute the smoothed loss
            smoothed_loss += avg_loss / (1 - self.beta**(i+1))   
            #Stop if the loss is exploding
            if i > 0 and smoothed_loss > 1000 * best_loss:
                break
            #Record the best loss
            if smoothed_loss < best_loss or i==0:
                best_loss = smoothed_loss
            #Store the values
            self.loss_list.append(smoothed_loss/len(train.dataset))
            self.log_lrs.append(np.log10(self.initial_lr))
            loss.backward()
            self.optimizer.step()
            #Update the lr for the next step
            self.initial_lr *= self.mult
            self.optimizer.param_groups[0]['lr'] = self.initial_lr
            
        
                
    def plot(self):
        plt.plot( self.log_lrs, self.loss_list,color='blue')
        plt.title('Plot b/w losses and lrs')
        plt.xlabel('LRs----> 1e')
        plt.ylabel('Loss----->')
        plt.show()




import numpy as np
from PIL import Image
from PIL import ImageFile   #ImageFile contains support for PIL to open and save images
ImageFile.LOAD_TRUNCATED_IMAGES=True  #If image is truncated(or corrupt) then also load it..
import torch

class Cloader:
    def __init__(self,image_path,targets,resize=False,transforms=False):
        self.image_path=image_path
        self.targets=targets
        self.resize=resize
        self.transforms=transforms
        
    def __len__(self):
        return len(self.image_path)
    
    def __getitem__(self,idx):
        image = Image.open(self.image_path[idx])
        targets = self.targets[idx]
        if self.resize is not None:
            image = image.resize(
                (self.resize[1], self.resize[0]), resample=Image.BILINEAR
            )
        image = np.array(image)
        if self.transforms is not None:
            augmented = self.transforms(image=image)
            image = augmented["image"]
        image = np.transpose(image, (2, 0, 1)).astype(np.float32)
        return {
            "image": torch.tensor(image, dtype=torch.float),
            "targets": torch.tensor(targets, dtype=torch.long),
        }




import torch

class Logger:
    #This has been specifically designed to store just the last epoch 
    #    (you can use it to save other epochs too..)
    def save(model,optimizer,scheduler,path):
        checkpoint = { 
        'model': model,
        'optimizer': optimizer,
        'scheduler': scheduler}
        torch.save(checkpoint, path)
    
    def load(path):
        checkpoint=torch.load(path)
        return checkpoint




import numpy as np
import pandas as pd
import torch
from sklearn import metrics
from tqdm import tqdm
from torch.autograd import Variable
from matplotlib import pyplot as plt
from IPython.display import clear_output
from matplotlib.animation import FuncAnimation
from matplotlib import style
plt.style.use('dark_background')
import sys
sys.tracebacklimit = 0


log=[]
train_list=[]
val_list=[]
class Trainer:
    @staticmethod
    def __init__():
        global epoch,val_loss,train_loss,log,train_list,val_list,lr
        log=[]
        train_list=[]
        val_list=[]

    def reset():
        global epoch,val_loss,train_loss,log,train_list,val_list
        epoch=0
        val_loss=0.0
        train_loss=0.0
        log=[]
        train_list=[]
        val_list=[]


    @staticmethod
    def train(model,data_loader,optimizer,device):
        global train_loss,lr,train_list
        model.train()
        optimizer.zero_grad()
        train_loss=0.0
        lr=optimizer.param_groups[0]['lr']
        tk = tqdm(data_loader, total=len(data_loader), position=0, leave=True)
        for idx,data in enumerate(tk):
            optimizer.zero_grad()
            for key, value in data.items():
                data[key] = value.to(device)
            outputs, loss = model(**data)
            loss.backward()
            optimizer.step()
            train_loss+=float((loss.data).item())/len(data_loader.dataset)
            tk.set_postfix(loss=train_loss)
        train_list.append(train_loss)

    @staticmethod
    def evaluate(model,data_loader,device,scheduler=None,metric=metrics.accuracy_score,plot=True):
        global epoch,val_loss,train_loss,log,train_list,val_list,lr
        final_predictions=[]
        targets=[]
        model.eval()
        val_loss=0.0
        with torch.no_grad():
            tk = tqdm(data_loader, total=len(data_loader), position=0, leave=True)
            for idx,data in enumerate(tk):
                for key, value in data.items():
                    data[key] = value.to(device)
                outputs, loss = model(**data)
                val_loss+=float((loss.data).item())/len(data_loader.dataset)
                final_predictions.append(outputs.cpu())
                targets.extend(data['targets'].detach().cpu().numpy())
        val_list.append(val_loss)
        pred = np.vstack((final_predictions))
        if metric==metrics.f1_score or metric==metrics.precision_score or metric==metrics.recall_score:
            if pred.shape[1]>1:
                preds=np.argmax(pred,axis=1)
            else:
                preds=np.round_(np.maximum(pred.ravel(),0))
            metric_score=metric(y_true=targets,y_pred=preds,average='micro') #Used micro as this is better option for imbalanced datas.
            if scheduler is not None:
                scheduler.step(metric_score)
            print(f'-------Validation metric_score:{metric_score}')
        elif metric==metrics.accuracy_score:
            if pred.shape[1]>1:
                preds=np.argmax(pred,axis=1)
            else:
                preds=np.round_(np.maximum(pred.ravel(),0))
            metric_score=metrics.accuracy_score(y_true=targets,y_pred=preds)
            if scheduler is not None:
                scheduler.step(metric_score)
            print(f'-------Validation accuracy_score:{metric_score}')
        elif metric==metrics.roc_auc_score:
            if pred.shape[1]>1:
                pred=torch.nn.functional.softmax(torch.from_numpy(pred))
                targets=onehot((pred.shape[0],pred.shape[1]),targets)
                metric_score=metrics.roc_auc_score(targets,pred,average='macro',multi_class='ovo')
            else:
                pred=torch.sigmoid(torch.from_numpy(pred))
                metric_score=metric(targets,pred,average='micro')
            if scheduler is not None:
                scheduler.step(metric_score)
            print(f'-------Validation roc_auc_score:{metric_score}')
        else:
            if pred.shape[1]>1:
                pred=torch.nn.functional.softmax(torch.from_numpy(pred))
                targets=onehot((pred.shape[0],pred.shape[1]),targets)
                metric_score=metric(targets,pred)
            else:
                pred=torch.sigmoid(torch.from_numpy(pred))
                metric_score=metric(targets,pred)
            if scheduler is not None:
                scheduler.step(metric_score)
            print(f'-------Validation metric_score:{metric_score}')
        log.append([train_loss,val_loss,metric_score,lr])
        clear_output(wait=True)
        print(pd.DataFrame(log,columns=['Train_Loss','Valid_Loss','Metric_Score','Current_LR']),flush=True)
        if plot:
            Trainer.Plotit()
        return metric_score
    
    @staticmethod
    def predict(model,data_loader,device):
        model.eval()
        final_predictions = []
        with torch.no_grad():
            tk = tqdm(data_loader, total=len(data_loader), position=0, leave=True)
            for idx,data in enumerate(tk):
                for key, value in data.items():
                    data[key] = value.to(device)
                outputs, loss = model(**data)
                final_predictions.append(outputs.data.cpu())
        pred=np.vstack(final_predictions)
        return pred
    
    def get_log():
        return pd.DataFrame(log,columns=['Train_Loss','Valid_Loss','Metric_Score','Current_LR'])

    def animate():
        plt.plot(np.arange(len(train_list)),train_list,color='blue' )
        plt.plot(np.arange(len(val_list)),val_list,color='orange' )
        plt.xlabel('Epochs------>')
        plt.ylabel('Losses------->')
        plt.title('Train and Val Loss Analysis')
        plt.legend(['Train_loss','Val_loss'])
        plt.gcf().autofmt_xdate()
        plt.tight_layout()

    def Plotit():
        ani = FuncAnimation(plt.gcf(), Trainer.animate(), 5000)
        plt.tight_layout()
        plt.show()

    def onehot(size, target):
        vec = torch.zeros(size, dtype=torch.float32)
        vec[target] = 1
        return vec







import torch
import numpy as np

class EarlyStop:
    def __init__(self, patience=7, mode="max", delta=0.0001):
        self.patience = patience
        self.counter = 0
        self.mode = mode
        self.best_score = None
        self.early_stop = False
        self.delta = delta
        if self.mode == "min":
            self.val_score = np.Inf
        else:
            self.val_score = -np.Inf
            
    def reset(self):        #Useful only when training for many folds.
        self.counter=0

    def __call__(self, epoch_score, model, model_path):
        if self.mode == "min":
            score = -1.0 * epoch_score
        else:
            score = np.copy(epoch_score)

        if self.best_score is None:
            self.best_score = score
            self.save(epoch_score, model, model_path)
        elif score < self.best_score + self.delta:
            self.counter += 1
            print(
                "EarlyStop count: {} out of {}".format(
                    self.counter, self.patience
                )
            )
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save(epoch_score, model, model_path)
            self.counter = 0

    def save(self, epoch_score, model, model_path):
        if epoch_score not in [-np.inf, np.inf, -np.nan, np.nan]:
            print(
                "Metric Validation score improved ({} --> {}). Saving model!".format(
                    self.val_score, epoch_score
                )
            )
            torch.save(model.state_dict(), model_path)
        self.val_score = epoch_score






from setuptools import setup, Extension
from setuptools import find_packages


with open("README.md", encoding="utf-8") as f:
    long_description = f.read()


if __name__ == "__main__":
    setup(
        name="pytorch_fanatics",
        version='0.2.3',
        description="A new library for Computer Vision",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="Harsh Mishra and Tanish Gupta",
        author_email="harshm17172612@gmail.com",
        url="https://github.com/MiHarsh/pytorch_fanatics",
        license="MIT License",
        packages=find_packages(),
        include_package_data=True,
        platforms=["linux", "unix"],
        python_requires=">3.5.2",
        install_requires=["scikit-learn>=0.22.1",],
    )
